__version__ = '0.0.1'

import sys
import requests
import gzip
import logging
import time
import json
import busrt
import pandas as pd

from datetime import datetime
from typing import Union
from io import StringIO, BytesIO

from evaics.client import HttpClient
from evaics.sdk import OID, FunctionFailed, pack, unpack

from busrt.rpc import Rpc as BusRpcClient

BUS_BULK_SIZE = 10_000


class HistoryDF:
    """
    Fetches data from EVA ICS v4 history databases

    When the primary module is imported, the method "history_df()" is
    automatically added to evaics.client.HttpClient
    """

    def __init__(self, client, csv: str = None):
        """
        Create HistoryDF object instance

        Args:
            client: HTTP client object

        All configuration methods of the class can be used in chains, e.g.
        HistoryDF(client).oid(
            'sensor:s1', status=False, value='s1').oid(
            'sensor:s2', status=False, value='s2').fill('1T').fetch()
        """
        self.client = client
        self.mlkit = None
        self.params = {'fill': '1S'}
        self.oid_map = {}

    def with_mlkit(self, mlkit: Union[bool, str]):
        """
        Set ML kit url/svc name

        Args:
            mlkit: True for the same URL as HMI, svc name or url for other
        """
        self.mlkit = mlkit
        return self

    def read_params_csv(self, f: str):
        """
        Read OID mapping from a CSV file

        CSV file must have the column "oid" and optional ones "status", "value"
        and "database"

        Args:
            f: file path or buffer
        """
        params = pd.read_csv(f)
        for _, row in params.iterrows():
            self.oid(row['oid'],
                     status=row.get('status'),
                     value=row.get('value'),
                     database=row.get('database'))
        return self

    def oid(self,
            oid: Union[OID, str],
            status=False,
            value=False,
            database=None,
            xopts=None):
        """
        Append OID for processing

        Args:
            oid: item OID (string or OID object)

        Optional:
            status: True to keep, a string to rename, False to drop
            value: same behavior as for keep
            database: db service to query data from (mlkit srv only)
            xopts: db service extra opts (mlkit srv only)
        """
        if status != status or \
                status == 'false' or status == '0' or status == 0:
            status = None
        elif status == 'true' or status == '1' or status == 1:
            status = True
        if value != value or \
                value == 'false' or value == '0' or value == 0:
            value = None
        elif value == 'true' or value == '1' or value == 1:
            value = True
        m = {'status': status, 'value': value}
        if database is not None:
            m['database'] = database
        if xopts is not None:
            m['xopts'] = xopts
        self.oid_map[oid] = m
        return self

    def t_start(self, t_start: Union[float, str, datetime]):
        """
        Specify the data frame start time

        Args:
            t_start: a float (timestamp), a string or a datetime object
        """
        if isinstance(t_start, datetime):
            t_start = t_start.timestamp()
        self.params['t_start'] = t_start
        return self

    def t_end(self, t_end: Union[float, str]):
        """
        Specify the data frame end time

        Args:
            t_start: a float (timestamp), a string or a datetime object
        """
        self.params['t_end'] = t_end
        return self

    def fill(self, fill: str):
        """
        Fill the data frame

        Args:
            fill: XN, where X - integer, N - fill type (S for seconds, T for
                  minutes, H for hours, D for days, W for weeks), e.g. 15T for
                  15 minutes. The values can be rounded to digits after comma
                  as XN:D, e.g 15T:2 - round to 2 digits after comma. The
                  default fill is 1S
        """
        self.params['fill'] = fill
        return self

    def limit(self, limit: int):
        """
        Limit the data frame rows to

        Args:
            limit: max number of rows
        """
        self.params['limit'] = limit
        return self

    def xopts(self, xopts: dict):
        """
        Extra database options

        Args:
            xopts: dict of extra options (refer to the EVA ICS database service
            documentation for more info)
        """
        if not self._is_mlkit_enabled():
            self.params['xopts'] = xopts
        for _, v in self.oid_map.items():
            v['xopts'] = xopts
        return self

    def database(self, database: str):
        """
        Specify a non-default database

        Args:
            database: database name (db service without eva.db. prefix)
        """
        if not self._is_mlkit_enabled():
            self.params['database'] = database
        for _, v in self.oid_map.items():
            v['database'] = database
        return self

    def _prepare_df_time(self, df, tz=None):
        if tz is None:
            tz = time.tzname[0]
        df['time'] = pd.to_datetime(df['time'], unit='s')
        df['time'] = df['time'].dt.tz_localize('UTC').dt.tz_convert(tz)

    def _get_mlkit(self):
        if self.mlkit is None:
            try:
                return self.client.mlkit
            except AttributeError:
                return None
        else:
            return self.mlkit

    def _is_mlkit_enabled(self):
        return self._get_mlkit() is not None

    def _get_ml_url(self):
        ml_url = self._get_mlkit()
        if ml_url is None:
            return None
        else:
            if ml_url is True:
                ml_url = self.client.url
            while ml_url.endswith('/'):
                ml_url = ml_url[:-1]
        return ml_url

    def _post(self, *args, accept_gzip=False, **kwargs):
        if not self.client.token:
            self.client.authenticate()
            refreshed = True
        else:
            refreshed = False
        while True:
            headers = {
                'x-auth-key': self.client.token,
            }
            if accept_gzip:
                headers['accept-encoding'] = 'gzip'
            req = requests.post(*args,
                                headers=headers,
                                timeout=self.client.timeout,
                                **kwargs)
            if req.status_code == 403 and req.text.endswith('(AUTH)'):
                if refreshed:
                    break
                else:
                    self.client.authenticate()
                    refreshed = True
            else:
                break
        return req

    def push(self, data, database='default'):
        """
        Push data

        Requires ML kit server

        Options:
            data: Pandas data frame, file object or file path
            database: database svc id (default: default)
        """
        if isinstance(self.client, BusRpcClient):
            raise RuntimeError('push via BUS/RT is not supported')
        else:
            ml_url = self._get_ml_url()
            if ml_url is None:
                raise RuntimeError('mlkit server not specified')
            params = {'database': database, 'oid_map': self.oid_map}
            files = {}
            if hasattr(data, 'seek'):
                files['file'] = data
            elif isinstance(data, str):
                files['file'] = open(data, 'r')
            elif isinstance(data, pd.DataFrame):
                buf = StringIO()
                data.to_csv(buf)
                buf.seek(0)
                files['file'] = buf
            else:
                raise ValueError('unsupported data kind')
            req = self._post(f'{ml_url}/ml/api/push.item.state_history_csv',
                             data={'params': json.dumps(params)},
                             files=files)
            if req.ok:
                return req.json()
            else:
                raise FunctionFailed(f'{req.status_code} {req.text}')

    def _prepare_mlkit_params(self, t_col):
        params = self.params.copy()
        params['oid_map'] = self.oid_map
        if t_col == 'drop':
            params['time_format'] = 'no'
        else:
            params['time_format'] = 'raw'
        return params

    def fetch(self, t_col: str = 'keep', tz: str = None):
        """
        Fetch data

        Optional:
            t_col: time column processing, "keep" - keep the column, "drop" -
                   drop the time column, "index" - keep and make the data frame
                   index
            tz: time zone

        Returns:
            a prepared Pandas DataFrame object
        """
        params = self.params.copy()
        if isinstance(self.client, BusRpcClient):
            params = self._prepare_mlkit_params(t_col)
            res = self.client.call(
                self._get_mlkit(),
                busrt.rpc.Request('Citem.state_history',
                                  pack(params))).wait_completed()
            cursor_id = unpack(res.get_payload(), raw=False)['u']
            cpl = pack(dict(u=cursor_id, c=BUS_BULK_SIZE))
            df = pd.DataFrame()
            while True:
                res = self.client.call(self._get_mlkit(),
                                       busrt.rpc.Request('NB',
                                                         cpl)).wait_completed()
                payload = res.get_payload()
                if payload:
                    if len(df.columns):
                        a = pd.read_csv(BytesIO(payload),
                                        header=None,
                                        names=df.columns)
                        df = pd.concat([df, a])
                        if len(a) < BUS_BULK_SIZE:
                            break
                    else:
                        df = pd.read_csv(BytesIO(payload))
                        if len(
                                df
                        ) < BUS_BULK_SIZE - 1:  # df rows (strings) + header (1 string)
                            break
                else:
                    break
            if t_col != 'drop':
                self._prepare_df_time(df, tz)
        else:
            ml_url = self._get_ml_url()
            if ml_url:
                params = self._prepare_mlkit_params(t_col)
                req = self._post(f'{ml_url}/ml/api/query.item.state_history',
                                 json=params,
                                 stream=True)
                if req.ok:
                    if req.headers.get('content-encoding') == 'gzip':
                        stream = gzip.GzipFile(fileobj=req.raw, mode="rb")
                    else:
                        stream = req.raw
                    df = pd.read_csv(stream)
                    if t_col != 'drop':
                        self._prepare_df_time(df, tz)
                else:
                    raise FunctionFailed(f'{req.status_code} {req.text}')
            else:
                params['i'] = list(self.oid_map)
                stats = self.client.call('item.state_history', params)
                for oid, proc in self.oid_map.items():
                    if not proc['status']:
                        del stats[f'{oid}/status']
                    if not proc['value']:
                        del stats[f'{oid}/value']
                cols = {}
                if t_col in ['keep', 'index']:
                    times = stats.pop('t')
                else:
                    del stats['t']
                    times = None
                df = pd.DataFrame(data=stats)
                if times is not None:
                    df.insert(0, 'time', times)
                for oid, proc in self.oid_map.items():
                    if not isinstance(proc['status'], bool):
                        cols[f'{oid}/status'] = proc['status']
                    if not isinstance(proc['value'], bool):
                        cols[f'{oid}/value'] = proc['value']
                if cols:
                    df = df.rename(columns=cols)
                if t_col != 'drop':
                    self._prepare_df_time(df, tz)
                return df
            if t_col == 'index':
                df = df.set_index('time')
        return df


def _history_df(self):
    return HistoryDF(self)


HttpClient.history_df = _history_df
BusRpcClient.history_df = _history_df



import pandas as pd
import numpy as np
import pickle

from types import SimpleNamespace
from tqdm import tqdm


def verify_array(arr: np.array, verbose=1, description=''):
    data = arr.ravel()
    rng = range(0, len(data))
    if verbose:
        rng = tqdm(rng)
        rng.set_description(description)
    for i in rng:
        if abs(data[i]) > 1:
            raise ValueError(
                'scaled data deviation is out of the recommended range')


def df_to_rolling_windows(df: pd.DataFrame,
                          window_size: int,
                          shift: int = 0,
                          limit: int = None,
                          verbose=1,
                          description=''):
    """
    Convert data frame to rolling windows numpy array

    Args:
        window_size: desired window size

    Optional:
        shift: shift data N rows ahead
        limit: limit the result data to N windows
        verbose: show data processing progress bar
        description: description for the progress bar if enabled

    Returns:
        processed numpy array
    """
    data = []
    n = len(df) - window_size - shift
    if limit is not None:
        l = limit + shift
        if l < n:
            n = l
    rng = range(shift, n)
    if verbose:
        rng = tqdm(rng)
        rng.set_description(description)
    for i in rng:
        data.append(df[i:i + window_size].to_numpy())
    return np.array(data)


def save_tf_session(path, model, data):
    """
    Saves model and session data

    Args:
        path: file path (without extension)
        model: TensorFlow model
        data: any object
    """
    model.save(f'{path}.h5')
    with open(f'{path}.dat', 'wb') as fh:
        pickle.dump(data, fh)


def load_tf_session(path):
    """
    Loads model and session data

    Args:
        path: file path (without extension)

    Returns:
        (model, data) tuple
    """
    from tensorflow import keras
    model = keras.models.load_model(f'{path}.h5')
    with open(f'{path}.dat', 'rb') as fh:
        data = pickle.load(fh)
    return (model, data)


class Regression:
    """
    A helper class for standard regression analysis

    All class configuration and load/save functions can be chained
    """

    def __init__(self, window_size: int = 30):
        """
        Optional:
            window_size: desired window size (for new models), default is 30
        """
        self._params = SimpleNamespace(window_size=window_size,
                                       scaled=False,
                                       scaler=None,
                                       y_cols=None,
                                       y_shift=None)
        self.dataframes = SimpleNamespace(x=None, y=None)
        self.data = SimpleNamespace(x=None, y=None)
        self.model = None

    def get_window_size(self):
        """
        Get window size

        Returns:
            window size
        """
        return self._params.window_size

    def load(self, path):
        """
        Loads model and session data

        Args:
            path: file path (without extension)
        """
        (self.model, self._params) = load_tf_session(path)
        return self

    def save(self, path):
        """
        Saves model and session data

        Args:
            path: file path (without extension)
        """
        self._require('model')
        save_tf_session(path, self.model, self._params)
        return self

    def with_scaler(self, scaler):
        """
        Set a custom scaler

        Must have the same API as MinMaxScaler

        Args:
            scaler: custom scaler object
        """
        self._params.scaler = scaler
        return self

    def with_standard_scaler(self):
        """
        Set scaler to MinMaxScaler
        """
        from sklearn.preprocessing import MinMaxScaler
        self._params.scaler = MinMaxScaler()
        return self

    def with_training_data(self,
                           data: pd.DataFrame,
                           y_cols: list = None,
                           y_shift: int = None):
        """
        Set training data

        Args:
            data: pandas.DataFrame

        Note: for the moment the supplied regression model supports a single
        Y-col only

        Optional:
            y_cols: a Y-column which is used as the training target, required
                    for the first training data set, can not be set later
            y_shift: shift Y-column N rows ahead, ignored for opaquie trainings
                    (the first set value is used)
        """
        self._require('scaler')
        if self._params.y_cols is None:
            if y_cols is None:
                raise RuntimeError('y_cols required for the first data set')
            self._params.y_cols = y_cols if isinstance(y_cols,
                                                       list) else [y_cols]
            if len(self._params.y_cols) != 1:
                raise ValueError('y_cols must have 1 column')
        elif self._params.y_cols is not None and y_cols is not None:
            raise RuntimeError('data model y_cols already set')
        if self._params.y_shift is None:
            self._params.y_shift = self._params.window_size if y_shift is None \
                    else y_shift
        if self._params.scaled:
            df = pd.DataFrame(self._params.scaler.transform(data),
                              columns=data.columns)
        else:
            df = pd.DataFrame(self._params.scaler.fit_transform(data),
                              columns=data.columns)
            self._params.scaled = True
        ydata = []
        for col in self._params.y_cols:
            ydata.append(pd.DataFrame(df.pop(col), columns=[col]))
        self.dataframes.x = df
        self.dataframes.y = pd.concat(ydata, axis=1)
        return self

    def with_prediction_data(self, data: pd.DataFrame):
        """
        Set prediction data

        The function automatically removes Y-colums from the data if found,
        send data.copy() if the data is required later

        Args:
            data: pandas.DataFrame
        """
        self._require('scaler')
        self._require('y_cols')
        for col in self._params.y_cols:
            data[col] = 0
        df = pd.DataFrame(self._params.scaler.transform(data),
                          columns=data.columns)
        for col in self._params.y_cols:
            del data[col]
            del df[col]
        self.dataframes.x = df
        self.dataframes.y = None
        return self

    def with_model(self, model):
        """
        Set a custom model

        The model must be compiled manually

        Args:
            model: TensorFlow model
        """
        self.model = model
        return self

    def with_standard_model(self,
                            dense_units=10,
                            loss='mse',
                            optimizer='adam',
                            **kwargs):
        """
        Set the standard model

        The standard model is a Sequential model, which contains:

        * two LSTM layers with number of units = window_size
        * N Dense units
        * one or several exit Dense units, depending on Y-cols set

        Opitonal:
            dense_units: dense units to set (default: 10)
            loss: loss function to compile the model (default: mse)
            optimizer: model optimizer (default: adam)
            kwargs: sent to model.compile as-is
        """
        self._require('df.x')
        self._require('df.y')
        from tensorflow.keras.models import Sequential
        from tensorflow.keras.layers import LSTM, Dense
        self.model = Sequential([
            LSTM(self._params.window_size,
                 input_shape=(self._params.window_size,
                              len(self.dataframes.x.columns)),
                 return_sequences=True),
            LSTM(self._params.window_size, return_sequences=False),
            Dense(dense_units * len(self.dataframes.y.columns)),
            Dense(len(self.dataframes.y.columns)),
        ])
        self.model.compile(loss=loss, optimizer=optimizer, **kwargs)
        return self

    def prepare_data(self, verbose=1):
        """
        Prepares the previously set data

        Optional:
            verbose: show verbose progress (default: 1)
        """
        self._require('df.x')
        if self.dataframes.y is None:
            self.data.y = None
        else:
            self.data.y = df_to_rolling_windows(
                self.dataframes.y,
                window_size=self._params.window_size,
                shift=self._params.y_shift,
                verbose=verbose,
                description='preparing y')
        self.data.x = df_to_rolling_windows(
            self.dataframes.x,
            window_size=self._params.window_size,
            limit=None if self.data.y is None else len(self.data.y),
            verbose=verbose,
            description='preparing x')
        return self

    def verify_prepared(self, verbose=1):
        """
        Verifies prepared data

        Optional:
            verbose: verbose

        Raises:
            ValueError: if scaled data deviation is out of the recommended range
        """
        self._require('data.x')
        if self.data.y is not None:
            verify_array(self.data.y,
                         verbose=verbose,
                         description='verifying y')
        verify_array(self.data.x, verbose=verbose, description='verifying x')
        return self

    def fit_model(self, epochs, shuffle=False, verbose=1, **kwargs):
        """
        Trains the model

        The data must be set and prepared

        Args:
            epochs: epochs to train

        Optional:
            shuffle: default: False
            verbose: default: 1
            kwargs: sent to model.fit as-is
        """
        self._require('data.x')
        self._require('data.y')
        self._require('model')
        return self.model.fit(self.data.x,
                              self.data.y,
                              epochs=epochs,
                              shuffle=shuffle,
                              verbose=verbose,
                              **kwargs)

    def predict(self, col_suffix='_predict', verbose=1, **kwargs):
        """
        Perform prediction using the pre-trained model

        The data must be set and prepared

        Optional:
            col_suffix: append a suffix to Y-cols, e.g. temp: temp_SFX, the
                        default is "_predict". Set to none to keep columns as
                        they are
            verbose: default: 1
        """
        self._require('data.x')
        self._require('model')
        self._require('y_cols')
        pred = self.model.predict(self.data.x, verbose=verbose, **kwargs)
        data = pd.concat([
            self.dataframes.x[:-self._params.window_size],
            pd.DataFrame(pred, columns=self._params.y_cols)
        ],
                         axis=1)
        df = pd.DataFrame(self._params.scaler.inverse_transform(data),
                          columns=data.columns)
        for col in data.columns:
            if col not in self._params.y_cols:
                del df[col]
        if col_suffix is not None:
            df.rename(columns={
                col: f'{col}{col_suffix}' for col in self._params.y_cols
            },
                      inplace=True)
        return df

    def _require(self, what):
        if what == 'scaler':
            if self._params.scaler is None:
                raise RuntimeError('scaler not set')
        elif what == 'model':
            if self.model is None:
                raise RuntimeError('model not set')
        elif what == 'df.x':
            if self.dataframes.x is None:
                raise RuntimeError('data not set (x)')
        elif what == 'df.y':
            if self.dataframes.y is None:
                raise RuntimeError('data not set (y)')
        elif what == 'data.x':
            if self.data.x is None:
                raise RuntimeError('data not prepared (x)')
        elif what == 'data.y':
            if self.dataframes.y is None:
                raise RuntimeError('data not prepared (y)')
        elif what == 'y_cols':
            if self._params.y_cols is None:
                raise RuntimeError('training data has been never set')
        else:
            raise RuntimeError(f'internal error, invalid attr {what}')



__version__ = '0.0.1'

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='evaics.ml',
    version=__version__,
    author='Bohemia Automation / Altertech',
    author_email='div@altertech.com',
    description='EVA ICS v4 Machine Learning Kit',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/eva-ics/eva-mlkit-client-python',
    packages=setuptools.find_packages(),
    license='Apache License 2.0',
    install_requires=[
        'requests>=2.26.0', 'pandas>=1.5.1', 'evaics>=0.2.5', 'tqdm>=4.45.0'
    ],
    classifiers=('Programming Language :: Python :: 3',
                 'License :: OSI Approved :: MIT License',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Scientific/Engineering :: Artificial Intelligence'),
)
